/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxbn.h"

#define VERSION_REV 16
#define VERSION_DATE "2017-05-18"

#if defined(_MSC_VER)
  #include "flxlibflxbn_export.h"
#else
  #define FLXLIBFLXBN_EXPORT
#endif


flxlibs_interface flx_iface;


using namespace flxbn;

extern "C" FLXLIBFLXBN_EXPORT void flxlib_init( FlxObjectReadBox* objReadBox, FlxData* dataBox ) {
  GlobalVar.slog(4) << "  FlxBN-library - Version " << VERSION_REV << " (" << VERSION_DATE << ")" << std::endl;
  flx_iface.set_boxes(objReadBox,&(dataBox->FunBox));
  // objects
    flx_iface.add_obj("bntest", new FlxObjReadBNtest()); 
    flx_iface.add_obj("bn_irv", new FlxObjReadBNirv()); 
    flx_iface.add_obj("bn_drv", new FlxObjReadBNdrv()); 
    flx_iface.add_obj("bn_detrel_and", new FlxObjReadBN_detrv_and()); 
    flx_iface.add_obj("bn_dec", new FlxObjReadBN_dec()); 
    flx_iface.add_obj("bn_utility", new FlxObjReadBNutility());
    flx_iface.add_obj("bn_observe", new FlxObjReadBNobserve());
    flx_iface.add_obj("bn_unobserve", new FlxObjReadBNunobserve());
    flx_iface.add_obj("bn_topo", new FlxObjReadBNtopo());
    flx_iface.add_obj("bn_smpmth", new FlxObjReadBNsmpmthd());
    flx_iface.add_obj("bn_elimorder_append", new FlxObjReadElimOrderAppend());
    flx_iface.add_obj("bn_train", new FlxObjReadBNtrain());
    flx_iface.add_obj("bn_train_radar", new FlxObjReadBNtrain_radar());
    flx_iface.add_obj("mtxconst_bn_condprobvec", new FlxObjReadMtxConstCondProbVec());
  // functions
    flx_iface.add_fun( "bn_prob",new FunReadBN_prob() );
    flx_iface.add_fun( "bn_condprob",new FunReadBN_condProb() );
    flx_iface.add_fun( "bn_exputility", new FunReadBN_expUtility() );
}

extern "C" FLXLIBFLXBN_EXPORT void flxlib_close() {
  flx_iface.free_memory();
}


//-------------------------------------------------------------------------------------------------


void FlxObjBNtest::task()
{
  sout() << "BN test" << std::endl;
}

FlxObjBase* FlxObjReadBNtest::read()
{
  read_optionalPara(false);
  return new FlxObjBNtest(get_doLog(), get_stream() );
}


// -------------------------- independent RV ----------------------------------------------------------------------

FlxObjReadBNirv::FlxObjReadBNirv()
: FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(ZERO,"flxbn::roundoff_correct"));
  ParaBox.insert("roundoff_correct", "flxbn::roundoff_correct" );
}

FlxObjBase* FlxObjReadBNirv::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  FlxMtxConstFun* ptable = NULL;
  try {
    reader->getChar(',',false);
      const tuint nstate = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<std::string> sv(nstate);
      for (tuint i=0;i<nstate;++i) {
	sv[i] = reader->getWord(true,false,true);
	if (i<nstate-1) {
	  reader->getChar(';',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(',',false);
    ptable = new FlxMtxConstFun(true);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjBNirv(get_doLog(),get_stream(),name,sv,ptable,get_optPara_FlxFunction("roundoff_correct"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNirv::read",1);
    delete name;
    if (ptable) delete ptable;
    throw;
  }
}

void FlxObjBNirv::task()
{
  const std::string nameStr = name->eval_word(true);
  tuint N = 0;
  const flxVec pt(data->ConstMtxBox.get_Vec(probVec->eval(),N),N);
  const tdouble roundoff_correct = rfc->cast2positive_or0(false);
  bnm.set_node(new bn_obj_irv(nameStr,stateVec,pt,roundoff_correct));
}

FlxObjBNirv::~FlxObjBNirv()
{
  delete name;
  delete probVec;
  delete rfc;
}

// -------------------------- dependent RV ----------------------------------------------------------------------

FlxObjReadBNdrv::FlxObjReadBNdrv()
: FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(ZERO,"flxbn::roundoff_correct"));
  ParaBox.insert("roundoff_correct", "flxbn::roundoff_correct" );
}

FlxObjBase* FlxObjReadBNdrv::read()
{
  reader->getChar('(',false);
  FlxString* name = new FlxString(false,false);
  FlxMtxConstFun* cptable = NULL;
  try {
    reader->getChar(',',false);
      const tuint nstate = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<std::string> sv(nstate);
      for (tuint i=0;i<nstate;++i) {
	sv[i] = reader->getWord(true,false,true);
	if (i<nstate-1) {
	  reader->getChar(';',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(',',false);
      const tuint ndepend = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<bn_rv_base*> dv(ndepend);
      for (tuint i=0;i<ndepend;++i) {
	const std::string dvname = reader->getWord(true,false);
	dv[i] = &(bnm.get_rv(dvname));
	if (dv[i]->outputRV==false) {
	  std::ostringstream ssV_2;
	  ssV_2 << "The random variable '" << dvname << "' is not an output-RV.";
	  throw FlxException("FlxObjReadBNdrv::read_1", ssV_2.str()); 
	}
	if (i<ndepend-1) {
	  reader->getChar(',',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(',',false);
    cptable = new FlxMtxConstFun(true);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjBNdrv(get_doLog(),get_stream(),name,sv,dv,cptable,get_optPara_FlxFunction("roundoff_correct"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNdrv::read_",1);
    delete name;
    if (cptable) delete cptable;
    throw;
  }
}

void FlxObjBNdrv::task()
{
  const std::string nameStr = name->eval_word(true);
  FlxSMtx* ptM = data->ConstMtxBox.get(cprobTab->eval(),true);
  const flxVec ptV(ptM->get_internalPtr(true),ptM->get_Ncoeff());
  const tdouble roundoff_correct = rfc->calc();
  bnm.set_node(new bn_obj_drv(nameStr,stateVec,dv,ptV,roundoff_correct));
}

FlxObjBNdrv::~FlxObjBNdrv()
{
  delete name;
  delete cprobTab;
  delete rfc;
}




FlxObjBase* FlxObjReadBN_detrv_and::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  try {
    const tuint nstate = 2;
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<std::string> sv(nstate);
      for (tuint i=0;i<nstate;++i) {
	sv[i] = reader->getWord(true,false,true);
	if (i<nstate-1) {
	  reader->getChar(';',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(',',false);
      const tuint ndepend = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<bn_rv_base*> dv(ndepend);
      for (tuint i=0;i<ndepend;++i) {
	const std::string dvname = reader->getWord(true,false);
	dv[i] = &(bnm.get_rv(dvname));
	if (dv[i]->outputRV==false) {
	  std::ostringstream ssV_2;
	  ssV_2 << "The random variable '" << dvname << "' is not an output-RV.";
	  throw FlxException("FlxObjReadBN_detrv_and::read_1", ssV_2.str()); 
	}
	if (i<ndepend-1) {
	  reader->getChar(',',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjBN_detrv_and(get_doLog(),get_stream(),name,sv,dv);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBN_detrv_and::read_2",1);
    delete name;
    throw;
  }
}

void FlxObjBN_detrv_and::task()
{
  const std::string nameStr = name->eval_word(true);
  bnm.set_node(new bn_obj_detrel_and(nameStr,stateVec,dv));
}

FlxObjBN_detrv_and::~FlxObjBN_detrv_and()
{
  delete name;
}



FlxObjBase* FlxObjReadBN_dec::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  try {
    reader->getChar(',',false);
      const tuint nstate = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<std::string> sv(nstate);
      for (tuint i=0;i<nstate;++i) {
	sv[i] = reader->getWord(true,false,true);
	if (i<nstate-1) {
	  reader->getChar(';',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjBN_dec(get_doLog(),get_stream(),name,sv);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBN_dec::read",1);
    delete name;
    throw;
  }
}

void FlxObjBN_dec::task()
{
  const std::string nameStr = name->eval_word(true);
  bnm.set_node(new bn_obj_dec(nameStr,stateVec));
}

FlxObjBN_dec::~FlxObjBN_dec()
{
  delete name;
}



FlxObjBase* FlxObjReadBNutility::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  FlxMtxConstFun* utVec = NULL;
  try {
    reader->getChar(',',false);
      const tuint ndepend = reader->get_UInt<tuint>();
    reader->getChar(',',false);
    reader->getChar('{',false);
      std::vector<bn_rv_base*> dv(ndepend);
      for (tuint i=0;i<ndepend;++i) {
	const std::string dvname = reader->getWord(true,false);
	dv[i] = &(bnm.get_rv(dvname));
	if (dv[i]->outputRV==false) {
	  std::ostringstream ssV_2;
	  ssV_2 << "The random variable '" << dvname << "' is not an output-RV.";
	  throw FlxException("FlxObjReadBNutility::read_1", ssV_2.str()); 
	}
	if (i<ndepend-1) {
	  reader->getChar(',',false);
	}
      }
    reader->getChar('}',false);
    reader->getChar(',',false);
    utVec = new FlxMtxConstFun(true);
    reader->getChar(')',false);
    read_optionalPara(false);
    return new FlxObjBNutility(get_doLog(),get_stream(),name,dv,utVec);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNutility::read_2",1);
    delete name;
    if (utVec) delete utVec;
    throw;
  }
}

void FlxObjBNutility::task()
{
  const std::string nameStr = name->eval_word(true);
  tuint N = 0;
  const flxVec utV = flxVec(data->ConstMtxBox.get_Vec(utVec->eval(),N),N);
  bnm.set_node(new bn_obj_utility(nameStr,dv,utV));
}

FlxObjBNutility::~FlxObjBNutility()
{
  delete name;
  delete utVec;
}



void FlxObjBNobserve::task()
{
  rv_obs_pair* obp = obsvData.compute();
  try {
    obp->observe();
    GlobalVar.slog(4) << "BN::observe ";
    obsvData.output(GlobalVar.slog(4));
    GlobalVar.slog(4) << std::endl;
    delete obp;
  } catch(FlxException &e) {
    FLXMSG("FlxObjBNobserve::task",1);
    delete obp;
    throw;
  }
}

FlxObjBNobserve::~FlxObjBNobserve()
{
  obsvData.deleteIT();
}

FlxObjBase* FlxObjReadBNobserve::read()
{
  BN_t_pair btp(reader,funReader);
  try {
    read_optionalPara(false);
    return new FlxObjBNobserve(get_doLog(),get_stream(),btp);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNobserve::read",1);
    btp.deleteIT();
    throw;
  }
}

void FlxObjBNunobserve::task()
{
  const std::string rvn = strV->eval_word(true,false);
  bn_rv_base& rv = bnm.get_rv(rvn);
  rv.unobserve();
  GlobalVar.slog(4) << "BN::unobserve " << rvn << std::endl;
}

FlxObjBNunobserve::~FlxObjBNunobserve()
{
  delete strV;
}

FlxObjBase* FlxObjReadBNunobserve::read()
{
  FlxString* strV = new FlxString(false,false);
  read_optionalPara(false);
  return new FlxObjBNunobserve(get_doLog(),get_stream(),strV);
}


void FlxObjBNtopo::task()
{
  if (clear) {
    bnm.topo_reset();
  } else {
    bnm.create_topo();
    if (is_verbose()) {
      bnm.output_topo(sout());
    }
  }
}

FlxObjReadBNtopo::FlxObjReadBNtopo()
: FlxObjReadOutputBase(true)
{
  // clear
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"bn::topoclear"));
    ParaBox.insert("topoclear", "bn::topoclear" );
}

FlxObjBase* FlxObjReadBNtopo::read()
{
  read_optionalPara(false);
  return new FlxObjBNtopo(get_doLog(),get_stream(),get_verbose(),get_optPara_bool("topoclear"));
}

BN_t_pair_state::BN_t_pair_state(ReadStream* reader, FlxFunctionReader* funReader)
: rv_stateS(NULL),rv_stateF(NULL)
{
  if (reader->whatIsNextChar()=='=') {
    reader->getChar('=');
    rv_stateS = new FlxString(false,false);
  } else {
    reader->getChar(':');
    rv_stateF = new FlxFunction(funReader,false);
  }
}

void BN_t_pair_state::deleteIT()
{
  if (rv_stateS) delete rv_stateS;
  if (rv_stateF) delete rv_stateF;
}

const tuint BN_t_pair_state::compute(const flxbn::bn_rv_base& rv) const
{
  if (rv_stateS) {
    return rv.get_stateID(rv_stateS->eval_word(true,false,true));
  } else {
    const tuint rvsid = rv_stateF->cast2tuintW0(false);
    if (rvsid>=rv.stateN) {
      std::ostringstream ssV;
      ssV << "StateID(" << rvsid << ") is not valid.";
      std::ostringstream ssV2;
      ssV2 << "range:=[0;" << rv.stateN-1 << "]";
      throw FlxException_NeglectInInteractive("BN_t_pair_state::compute", ssV.str() );
    }
    return rvsid;
  }
}

void BN_t_pair_state::output(std::ostream& sout) const
{
  if (rv_stateS) {
    const std::string evs = rv_stateS->eval_word(true,false,true);
    sout << "=" << evs;
  } else {
    tuint sid = rv_stateF->cast2tuintW0(false);
    sout << ":" << sid;
  }
}

void BN_t_pair_state::write(std::ostream& sout) const
{
  if (rv_stateS) {
    sout << "=" << rv_stateS->write();
  } else {
    sout << ":" << rv_stateF->write();
  }
}

const bool BN_t_pair_state::search_circref(FlxFunction* fcr)
{
  if (rv_stateS) {
    return rv_stateS->search_circref(fcr);
  } else {
    return rv_stateF->search_circref(fcr);
  }
}

BN_t_pair::BN_t_pair(ReadStream* reader, FlxFunctionReader* funReader)
:rv_name(NULL)
{
  try {
    rv_name = new FlxString(false,false);
    if (reader->whatIsNextChar()=='{') {
      reader->getChar('{');
      while (true) {
	stateVec.push_back(BN_t_pair_state(reader,funReader));
	if (reader->whatIsNextChar()!=';') break;
	reader->getChar(';');
      };
      if (stateVec.empty()) {
	std::ostringstream ssV;
	ssV << "At least one state has to be observed.";
	throw FlxException_NeglectInInteractive("t_pair::t_pair", ssV.str() );
      }
      reader->getChar('}');
    } else {
      stateVec.push_back(BN_t_pair_state(reader,funReader));
    }
  } catch (FlxException &e) {
    FLXMSG("BN_t_pair::BN_t_pair",1);
    deleteIT();
    throw;
  }
}

void BN_t_pair::deleteIT()
{
  if (rv_name) delete rv_name;
  for (tuint i=0;i<stateVec.size();++i) {
    stateVec[i].deleteIT();
  }
}

rv_obs_pair* BN_t_pair::compute() const
{
  bn_rv_base* const rv = &(bnm.get_rv(rv_name->eval_word(true)));
  std::vector<tuint> observedStates;
  for (tuint i=0;i<stateVec.size();++i) {
    observedStates.push_back(stateVec[i].compute(*rv));
  }
  return new rv_obs_pair(rv,observedStates);
}

void BN_t_pair::output(std::ostream& sout) const
{
  sout << rv_name->eval_word(true);
  if (stateVec.size()>1) {
    sout << '{';
    for (tuint i=0;i<stateVec.size();++i) {
      stateVec[i].output(sout);
      if (i<stateVec.size()-1) sout << ';';
    }
    sout << '}';
  } else {
     stateVec[0].output(sout);
  }
}

void BN_t_pair::write(std::ostream& sout) const
{
  sout << rv_name->write();
  if (stateVec.size()>1) {
    sout << '{';
    for (tuint i=0;i<stateVec.size();++i) {
      stateVec[i].write(sout);
      if (i<stateVec.size()-1) sout << ';';
    }
    sout << '}';
  } else {
     stateVec[0].write(sout);
  }
}

const bool BN_t_pair::search_circref(FlxFunction* fcr)
{
  if (rv_name->search_circref(fcr)) return true;
  for (tuint i=0;i<stateVec.size();++i) {
    if (stateVec[i].search_circref(fcr)) return true;
  }
  return false;
}

FunBase* FunReadBN_condProb::read(bool errSerious)
{
  std::vector<BN_t_pair> il;
  try {
    bool started = false;
    while (reader->whatIsNextChar()!=')') {
      if (started) reader->getChar(',');
      else started = true;
      il.push_back(BN_t_pair(reader,funReader));
    }
    return new FunBN_condProb(il);
  } catch (FlxException &e) {
    FLXMSG("FunReadBN_condProb::read",1);
    for (tuint i=0;i<il.size();++i) {
      il[i].deleteIT();
    }
    throw;
  }
}

FunBN_condProb::~FunBN_condProb()
{
  for (tuint i=0;i<il.size();++i) {
    il[i].deleteIT();
  }
}

const tdouble FunBN_condProb::calc()
{
  std::vector<rv_obs_pair*> ilt;
  try {
    bnm.create_topo();
    // evaluate list
      std::ostringstream ssV;
      ssV << "Pr(";
      const tuint S = il.size();
      for (tuint i=0;i<S;++i) {
	if (i>0) ssV << ",";
	ilt.push_back(il[i].compute());
	il[i].output(ssV);
      }
      ssV << "|";
      bnm.output_observed(ssV);
      ssV << ")=";
    const tdouble cpV = bnm.calc_condProb(ilt);
    GlobalVar.slog(4) << ssV.str() << GlobalVar.Double2String(cpV) << std::endl;
    for (tuint i=0;i<ilt.size();++i) {
      delete ilt[i];
    }
    return cpV;
  } catch (FlxException &e) {
    FLXMSG("FunBN_condProb::calc",1);
    for (tuint i=0;i<ilt.size();++i) {
      delete ilt[i];
    }
    throw;
  }
}

const bool FunBN_condProb::search_circref(FlxFunction* fcr)
{
  for (tuint i=0;i<il.size();++i) {
    if (il[i].search_circref(fcr)) return true;
  }
  return false;
}

const std::string FunBN_condProb::write()
{
  std::ostringstream ssV;
  ssV << "bn_condprob(";
  for (tuint i=0;i<il.size();++i) {
    if (i>0) ssV << ",";
    il[i].write(ssV);
  }
  ssV << ")";
  return ssV.str();
}

const bool FunBN_condProb::dependOn_Const(const tdouble* const thenumber)
{
  throw FlxException_NotImplemented("FunBN_condProb::dependOn_Const");
}

void FlxObjMtxConstCondProbVec::task()
{
  bnm.create_topo();
  bn_rv_base& rvb = bnm.get_rv(rvn->eval_word(true));
  flxVec tv = bnm.calc_condProbVec(rvb);
  std::ostringstream ssV;
  ssV << "Pr(" << rvb.name << "|";
  bnm.output_observed(ssV);
  ssV << ")=";
  ssV << tv << std::endl;
  tuint N = tv.get_N();
  tdouble s = ZERO;
  for (tuint i=0;i<N;++i) s+=tv[i];
  ssV << "  sum-1=" << GlobalVar.Double2String(s-ONE);
  if (correct && fabs(s-ONE)>GlobalVar.TOL()) {
    ssV << "  (f=" << GlobalVar.Double2String(ONE/s) << ")" <<  std::endl;
    const tdouble f = ONE/s;
    tv *= f; 
    ssV << "  correct probability-vector ...: " << tv << std::endl;
    s = ZERO;
    for (tuint i=0;i<N;++i) s+=tv[i];
    ssV << "  again: sum-1=" << GlobalVar.Double2String(s-ONE) << std::endl;
  } else {
    ssV << std::endl;
  }
  GlobalVar.slog(4) << ssV.str();
  flxVec resV(data->ConstMtxBox.get_Vec(mcn_target->eval(true),N),N);
  resV = tv;
}

const tdouble FunBN_prob::calc()
{
  bnm.create_topo();
  // evaluate list
    std::ostringstream ssV;
    ssV << "Pr(";
    bnm.output_observed(ssV);
    ssV << ")=";
//   if (is_verbose()) {
    GlobalVar.slog(4) << ssV.str();
//   } else {
//     sout() << "Pr()=";
//   } 
  const tdouble pV = bnm.calc_probObservation();
  GlobalVar.slog(4) << GlobalVar.Double2String(pV) << std::endl;
  return pV;
}

const std::string FunBN_prob::write()
{
  return "bn_prob()";
}

FunBase* FunReadBN_prob::read(bool errSerious)
{
  return new FunBN_prob();
}

const tdouble FunBN_expUtility::calc()
{
  bnm.create_topo();
  std::ostringstream ssV;
  ssV << "E[Utility|";
  bnm.output_observed(ssV);
  ssV << "]=";
  const tdouble eut = bnm.calc_expUtility();
  GlobalVar.slog(4) << ssV.str() << GlobalVar.Double2String(eut) << std::endl;
  return eut;
}

const std::string FunBN_expUtility::write()
{
  return "bn_exputility()";
}

FunBase* FunReadBN_expUtility::read(bool errSerious)
{
  return new FunBN_expUtility();
}

void FlxObjBNsmpmthd::task()
{
  const tuint smp_mthd = mthF->cast2tuintW0(false);
  switch (smp_mthd) {
    case 0:
      break;
    case 1:
      bnm.set_smp_mthd_2_N(NF->cast2tuint(false));
      bnm.set_smp_mthd_2_eP(eqProb);
      break;
    case 2:
      break;
  }
  bnm.set_smp_mthd(smp_mthd);
}

FlxObjBNsmpmthd::~FlxObjBNsmpmthd()
{
  delete mthF;
  delete NF;
}

FlxObjReadBNsmpmthd::FlxObjReadBNsmpmthd()
: FlxObjReadOutputBase(true)
{
  // N
    AllDefParaBox->insert(new FlxOptionalParaFun(1e5,"bn::n"));
    ParaBox.insert("n", "bn::n" );
  // eqProb
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"bn::eqprob"));
    ParaBox.insert("eqprob", "bn::eqprob" );
}

FlxObjBase* FlxObjReadBNsmpmthd::read()
{
  reader->getChar('=');
  FlxFunction *mthF = new FlxFunction(funReader,false);
  try {
    read_optionalPara(false);
    return new FlxObjBNsmpmthd(get_doLog(),get_stream(),mthF,get_optPara_FlxFunction("n"),get_optPara_bool("eqprob"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNsmpmthd::read",1);
    delete mthF;
    throw;
  }
}

void FlxObjBNelimOrderAppend::task()
{
  const std::string rvs = rvStr->eval_word(true);
  bnm.append_elimOrderVar(&(bnm.get_rv(rvs)));
}

FlxObjBNelimOrderAppend::~FlxObjBNelimOrderAppend()
{
  delete rvStr;
}

FlxObjBase* FlxObjReadElimOrderAppend::read()
{
  reader->getChar('(');
  FlxString* name = new FlxString(false,false);
  try {
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjBNelimOrderAppend(get_doLog(),get_stream(),name);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadElimOrderAppend::read",1);
    delete name;
    throw;
  }
}


void FlxObjBNtrain::task()
{
  const std::string nameV = name->eval_word(true);
  tuint N = 0;
  flxVec ivalsV(data->ConstMtxBox.get_Vec(ivals->eval(),N),N);
  N = 0;
  flxVec ptableV(data->ConstMtxBox.get_Vec(ptable->eval(),N),N);
  tuint Nr=0,Nc=0;
  flxVec cptableV(data->ConstMtxBox.get_Mtx(cptable->eval(),Nr,Nc),Nr*Nc);
  Nr=0; Nc=0;
  flxVec rcptableV(data->ConstMtxBox.get_Mtx(rcptable->eval(),Nr,Nc),Nr*Nc);
  N = 0;
  flxVec sivalsV(data->ConstMtxBox.get_Vec(sivals->eval(),N),N);
  const tuint NstagesV  = Nstages->cast2tuint(false);
  const tdouble roundoff_correct = rfc->calc();
  bnm.set_node(new bn_obj_train(nameV,ivalsV,ptableV,cptableV,rcptableV,sivalsV,NstagesV,roundoff_correct));

}

FlxObjReadBNtrain::FlxObjReadBNtrain()
: FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(ZERO,"flxbn::roundoff_correct"));
  ParaBox.insert("roundoff_correct", "flxbn::roundoff_correct" );
}

FlxObjBNtrain::~FlxObjBNtrain()
{
  delete name;
  delete ivals;
  delete ptable;
  delete cptable;
  delete rcptable;
  delete sivals;
  delete Nstages;
  delete rfc;
}

FlxObjBase* FlxObjReadBNtrain::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  FlxMtxConstFun* ivals = NULL;
  FlxMtxConstFun* ptable = NULL;
  FlxMtxConstFun* cptable = NULL;
  FlxMtxConstFun* rcptable = NULL;
  FlxMtxConstFun* sivals = NULL;
  FlxFunction* nstages = NULL;
  try {
    reader->getChar(',',false);
    ivals = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    ptable = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    cptable = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    rcptable = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    sivals = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    nstages = new FlxFunction(funReader,false);
  reader->getChar(')',false);
  read_optionalPara(false);
  return new FlxObjBNtrain(get_doLog(),get_stream(),name,ivals,ptable,cptable,rcptable,sivals,nstages,get_optPara_FlxFunction("roundoff_correct"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNtrain::read",1);
    if (name) delete name;
    if (ivals) delete ivals;
    if (ptable) delete ptable;
    if (cptable) delete cptable;
    if (rcptable) delete rcptable;
    if (sivals) delete sivals;
    if (nstages) delete nstages;
    throw;
    return NULL;
  }
}

FlxObjReadBNtrain_radar::FlxObjReadBNtrain_radar()
: FlxObjReadOutputBase(true)
{
  AllDefParaBox->insert(new FlxOptionalParaFun(ZERO,"flxbn::roundoff_correct"));
  ParaBox.insert("roundoff_correct", "flxbn::roundoff_correct" );
}

FlxObjBase* FlxObjReadBNtrain_radar::read()
{
  reader->getChar('(',false);
    FlxString* name = new FlxString(false,false);
  FlxMtxConstFun* ivals = NULL;
  FlxMtxConstFun* cpt_RonA = NULL;
  FlxMtxConstFun* sivals = NULL;
  FlxFunction* nstages = NULL;
  try {
    reader->getChar(',',false);
    ivals = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    cpt_RonA = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    sivals = new FlxMtxConstFun(true);
    reader->getChar(',',false);
    nstages = new FlxFunction(funReader,false);
  reader->getChar(')',false);
  read_optionalPara(false);
  return new FlxObjBNtrain_radar(get_doLog(),get_stream(),name,ivals,cpt_RonA,sivals,nstages,get_optPara_FlxFunction("roundoff_correct"));
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadBNtrain_radar::read",1);
    if (name) delete name;
    if (ivals) delete ivals;
    if (cpt_RonA) delete cpt_RonA;
    if (sivals) delete sivals;
    if (nstages) delete nstages;
    throw;
    return NULL;
  }
}

void FlxObjBNtrain_radar::task()
{
  const std::string nameV = name->eval_word(true);
  tuint N = 0;
  flxVec ivalsV(data->ConstMtxBox.get_Vec(ivals->eval(),N),N);
  tuint Nr=0,Nc=0;
  flxVec cptableRonAV(data->ConstMtxBox.get_Mtx(cpt_RonA->eval(),Nr,Nc),Nr*Nc);
  N = 0;
  flxVec sivalsV(data->ConstMtxBox.get_Vec(sivals->eval(),N),N);
  const tuint NstagesV  = Nstages->cast2tuint(false);
  const tdouble roundoff_correct = rfc->calc();
  bnm.set_node(new bn_obj_train_radar(nameV,ivalsV,cptableRonAV,sivalsV,NstagesV,roundoff_correct));

}

FlxObjBNtrain_radar::~FlxObjBNtrain_radar()
{
  delete name;
  delete ivals;
  delete cpt_RonA;
  delete sivals;
  delete Nstages;
  delete rfc;
}

FlxObjBase* FlxObjReadMtxConstCondProbVec::read()
{
  FlxString* mcn_target = new FlxString(false,false);
  FlxString* rvn = NULL;
  try {
    reader->getChar('(');
    rvn = new FlxString(false,false);
    bool correct = true;
    if (reader->whatIsNextChar()==',') {
      reader->getChar(',');
      correct = reader->getBool(false);
    }
    reader->getChar(')');
    read_optionalPara(false);
    return new FlxObjMtxConstCondProbVec(get_doLog(),mcn_target,rvn,correct);
  } catch (FlxException &e) {
    FLXMSG("FlxObjReadMtxConstCondProbVec::read",1);
    delete mcn_target;
    if (rvn) delete rvn;
    throw;
  }
}

