/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef flxbn_flxbn_H
#define flxbn_flxbn_H



#include "flxlibs_interface.h"
#include "flxbn_obj.h"


namespace flxbn {

// -------------------------- FlxObjBNtest ----------------------------------------------------------------------

/**
* @brief object class: BN test class
*/
class FlxObjBNtest : public FlxObjOutputBase {
  private:
    void task();
  public:
    FlxObjBNtest ( bool dolog, std::string ostreamV ) : FlxObjOutputBase(dolog,ostreamV) {}
};

/**
* @brief object read class: for FlxObjBNtest
*/
class FlxObjReadBNtest : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNtest() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- independent RV ----------------------------------------------------------------------

class FlxObjBNirv : public FlxObjOutputBase {
  private:
    FlxString* name;
    const std::vector<std::string> stateVec;
    FlxMtxConstFun* probVec;
    FlxFunction* rfc;
    void task();
  public:
    FlxObjBNirv( const bool dolog, std::string ostreamV, FlxString* name, const std::vector<std::string>& stateVec, FlxMtxConstFun* probVec, FlxFunction* rfc) : FlxObjOutputBase(dolog,ostreamV), name(name), stateVec(stateVec), probVec(probVec), rfc(rfc) {}
    ~FlxObjBNirv();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNirv : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNirv();
    FlxObjBase* read ();
};


// -------------------------- dependent RV ----------------------------------------------------------------------

class bn_rv_base;
class FlxObjBNdrv : public FlxObjOutputBase {
  private:
    FlxString* name;
    const std::vector<std::string> stateVec;
    const std::vector<bn_rv_base*> dv;
    FlxMtxConstFun* cprobTab;
    FlxFunction* rfc;
    void task();
  public:
    FlxObjBNdrv( const bool dolog, std::string ostreamV, FlxString* name, const std::vector<std::string>& stateVec, const std::vector<bn_rv_base*>& dv, FlxMtxConstFun* cprobTab, FlxFunction* rfc) : FlxObjOutputBase(dolog,ostreamV), name(name), stateVec(stateVec), dv(dv), cprobTab(cprobTab), rfc(rfc) {}
    ~FlxObjBNdrv();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNdrv : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNdrv();
    FlxObjBase* read ();
};


// -------------------------- deterministic RV (and) ----------------------------------------------------------------------

class FlxObjBN_detrv_and : public FlxObjOutputBase {
  private:
    FlxString* name;
    const std::vector<std::string> stateVec;
    const std::vector<bn_rv_base*> dv;
    void task();
  public:
    FlxObjBN_detrv_and( const bool dolog, std::string ostreamV, FlxString* name, const std::vector<std::string>& stateVec, const std::vector<bn_rv_base*>& dv) : FlxObjOutputBase(dolog,ostreamV), name(name), stateVec(stateVec), dv(dv) {}
    ~FlxObjBN_detrv_and();
};

/**
* @brief object read class: 
*/
class FlxObjReadBN_detrv_and : public FlxObjReadOutputBase {
  public:
    FlxObjReadBN_detrv_and() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- decission node ----------------------------------------------------------------------


class FlxObjBN_dec : public FlxObjOutputBase {
  private:
    FlxString* name;
    const std::vector<std::string> stateVec;
    void task();
  public:
    FlxObjBN_dec( const bool dolog, std::string ostreamV, FlxString* name, const std::vector<std::string>& stateVec) : FlxObjOutputBase(dolog,ostreamV), name(name), stateVec(stateVec) {}
    ~FlxObjBN_dec();
};

/**
* @brief object read class: 
*/
class FlxObjReadBN_dec : public FlxObjReadOutputBase {
  public:
    FlxObjReadBN_dec() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- utility ----------------------------------------------------------------------

class FlxObjBNutility : public FlxObjOutputBase {
  private:
    FlxString* name;
    const std::vector<bn_rv_base*> dv;
    FlxMtxConstFun* utVec;
    void task();
  public:
    FlxObjBNutility( const bool dolog, std::string ostreamV, FlxString* name, const std::vector<bn_rv_base*> dv, FlxMtxConstFun* utVec) : FlxObjOutputBase(dolog,ostreamV), name(name), dv(dv), utVec(utVec) {}
    ~FlxObjBNutility();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNutility : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNutility() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- observe ----------------------------------------------------------------------

class bn_rv_base;

class BN_t_pair_state {
  private:
    FlxString* rv_stateS;
    FlxFunction* rv_stateF;
  public:
    BN_t_pair_state(ReadStream* reader, FlxFunctionReader* funReader);
    void deleteIT();
    const tuint compute(const bn_rv_base& rv) const;
    void output(std::ostream &sout) const;
    void write(std::ostream &sout) const;
    const bool search_circref(FlxFunction* fcr);
};

class rv_obs_pair;

class BN_t_pair {
  private:
    FlxString* rv_name;
    std::vector<BN_t_pair_state> stateVec;
  public:
    BN_t_pair(ReadStream* reader, FlxFunctionReader* funReader);
    void deleteIT();
    /**
    * @brief evaluate the functions ...
    * @note you have to free the memory !!!
    */
    rv_obs_pair* compute() const;
    void output(std::ostream &sout) const;
    void write(std::ostream &sout) const;
    const bool search_circref(FlxFunction* fcr);
};

/**
* @brief object class: make an observation in the BN
*/
class FlxObjBNobserve : public FlxObjOutputBase {
  private:
    BN_t_pair obsvData;
    void task();
  public:
    FlxObjBNobserve ( const bool dolog, std::string ostreamV, BN_t_pair& obsvData ) : FlxObjOutputBase(dolog,ostreamV), obsvData(obsvData) {}
    ~FlxObjBNobserve();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNobserve : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNobserve() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};

// -------------------------- unobserve ----------------------------------------------------------------------

/**
* @brief object class: undo an observation in the BN
*/
class FlxObjBNunobserve : public FlxObjOutputBase {
  private:
    FlxString* strV;
    void task();
  public:
    FlxObjBNunobserve ( const bool dolog, std::string ostreamV, FlxString* strV) : FlxObjOutputBase(dolog,ostreamV), strV(strV) {}
    ~FlxObjBNunobserve();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNunobserve : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNunobserve() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
};


// -------------------------- topology ----------------------------------------------------------------------

/**
* @brief object class: create topology of the BN
*/
class FlxObjBNtopo : public FlxObjOutputBase {
  private:
    const bool clear;
    void task();
  public:
    FlxObjBNtopo ( const bool dolog, std::string ostreamV, const bool verboseV, const bool clear) : FlxObjOutputBase(dolog,ostreamV,verboseV), clear(clear) {}
};

/**
* @brief object read class: 
*/
class FlxObjReadBNtopo : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNtopo();
    FlxObjBase* read ();
};


// -------------------------- condProb ----------------------------------------------------------------------

class FunBN_condProb : public FunBase {
  private:
    std::vector<BN_t_pair> il;
  public:
    FunBN_condProb(const std::vector<BN_t_pair>& il) : il(il) {}
    ~FunBN_condProb();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

/**
* @brief arithmetic read class of FunFEMelInfo
*/
class FunReadBN_condProb : public FunReadFunBase, public FlxReaderBase2 {
  public:
    FunReadBN_condProb() : FunReadFunBase(true) {}
    FunBase* read ( bool errSerious );
};

// -------------------------- condProbTable ----------------------------------------------------------------------

class FlxObjMtxConstCondProbVec : public FlxObjBase {
  protected:
    FlxString* mcn_target;		// where to store
    FlxString* rvn;
    const bool correct;
    
    void task();
  public:
    FlxObjMtxConstCondProbVec ( bool dolog, FlxString* mcn_target, FlxString* rvn, const bool correct ) : FlxObjBase(dolog), mcn_target(mcn_target), rvn(rvn), correct(correct) {}
    virtual ~FlxObjMtxConstCondProbVec() { delete mcn_target; delete rvn; }
};

class FlxObjReadMtxConstCondProbVec : public FlxObjReadBase {
  public:
    FlxObjBase* read ();
};

// -------------------------- Prob ----------------------------------------------------------------------

class FunBN_prob : public FunBase {
  public:
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write();
    const tuint precedence() { return 0; }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

/**
* @brief arithmetic read class of FunFEMelInfo
*/
class FunReadBN_prob : public FunReadFunBase {
  public:
    FunReadBN_prob() : FunReadFunBase(true) {}
    FunBase* read ( bool errSerious );
};


// -------------------------- expected utility ----------------------------------------------------------------------

class FunBN_expUtility : public FunBase {
  public:
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write();
    const tuint precedence() { return 0; }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

/**
* @brief arithmetic read class of FunFEMelInfo
*/
class FunReadBN_expUtility : public FunReadFunBase {
  public:
    FunReadBN_expUtility() : FunReadFunBase(true) {}
    FunBase* read ( bool errSerious );
};

// -------------------------- define sampling method ----------------------------------------------------------------------

/**
* @brief object class: chose a sampling method to compute probabilities in the BN
*/
class FlxObjBNsmpmthd : public FlxObjOutputBase {
  private:
    FlxFunction* NF;
    FlxFunction* mthF;
    const bool eqProb;
    void task();
  public:
    FlxObjBNsmpmthd ( const bool dolog, std::string ostreamV, FlxFunction* mthF, FlxFunction* NF, const bool eqProb) : FlxObjOutputBase(dolog,ostreamV),NF(NF), mthF(mthF), eqProb(eqProb) {}
    ~FlxObjBNsmpmthd();
};

/**
* @brief object read class: 
*/
class FlxObjReadBNsmpmthd : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNsmpmthd();
    FlxObjBase* read ();
};


// -------------------------- define elimination order ----------------------------------------------------------------------

/**
* @brief object class: chose a sampling method to compute probabilities in the BN
*/
class FlxObjBNelimOrderAppend : public FlxObjOutputBase {
  private:
    FlxString* rvStr;
    void task();
  public:
    FlxObjBNelimOrderAppend ( const bool dolog, std::string ostreamV, FlxString* rvStr) : FlxObjOutputBase(dolog,ostreamV), rvStr(rvStr) {}
    ~FlxObjBNelimOrderAppend();
};

class FlxObjReadElimOrderAppend : public FlxObjReadOutputBase {
  public:
    FlxObjReadElimOrderAppend() : FlxObjReadOutputBase(true) {}
    FlxObjBase* read ();
}; 


// -------------------------- temporal rain model ----------------------------------------------------------------------

/**
* @brief object class: temporal rain model
*/
class FlxObjBNtrain : public FlxObjOutputBase {
  private:
    FlxString* name;
    FlxMtxConstFun* ivals;
    FlxMtxConstFun* ptable;
    FlxMtxConstFun* cptable;
    FlxMtxConstFun* rcptable;
    FlxMtxConstFun* sivals;
    FlxFunction* Nstages;
    FlxFunction* rfc;
    void task();
  public:
    FlxObjBNtrain ( const bool dolog, std::string ostreamV, FlxString* name, FlxMtxConstFun* ivals, FlxMtxConstFun* ptable, FlxMtxConstFun* cptable, FlxMtxConstFun* rcptable, FlxMtxConstFun* sivals, FlxFunction* Nstages, FlxFunction* rfc) : FlxObjOutputBase(dolog,ostreamV), name(name),ivals(ivals),ptable(ptable),cptable(cptable),rcptable(rcptable),sivals(sivals),Nstages(Nstages),rfc(rfc) {}
    ~FlxObjBNtrain();
};

class FlxObjReadBNtrain : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNtrain();
    FlxObjBase* read ();
}; 

// -------------------------- temporal rain model (no correlation) ----------------------------------------------------------------------

/**
* @brief object class: temporal rain model
*/
class FlxObjBNtrain_radar : public FlxObjOutputBase {
  private:
    FlxString* name;
    FlxMtxConstFun* ivals;
    FlxMtxConstFun* cpt_RonA;
    FlxMtxConstFun* sivals;
    FlxFunction* Nstages;
    FlxFunction* rfc;
    void task();
  public:
    FlxObjBNtrain_radar ( const bool dolog, std::string ostreamV, FlxString* name, FlxMtxConstFun* ivals, FlxMtxConstFun* cpt_RonA, FlxMtxConstFun* sivals, FlxFunction* Nstages, FlxFunction* rfc) : FlxObjOutputBase(dolog,ostreamV), name(name),ivals(ivals),cpt_RonA(cpt_RonA),sivals(sivals),Nstages(Nstages),rfc(rfc) {}
    ~FlxObjBNtrain_radar();
};

class FlxObjReadBNtrain_radar : public FlxObjReadOutputBase {
  public:
    FlxObjReadBNtrain_radar();    
    FlxObjBase* read ();
}; 






} // end of namespace 'flxbn'


#endif // flxbn_flxbn_H