/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2017 Wolfgang Betz
 *
 * Fesslix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Fesslix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Fesslix.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef flxbn_flxbn_obj_H
#define flxbn_flxbn_obj_H

#include "flxbn_manager.h"

namespace flxbn {


class bn_obj_irv : public bn_node {
  private:
    const std::vector<std::string> states;
    flxVec table;
    bn_rv rv;
  public:
    bn_obj_irv(const std::string name, const std::vector<std::string>& statesV, const flxVec& tableV, const tdouble roundoff_correct);
};

class bn_obj_drv : public bn_node {
  private:
    const std::vector<std::string> states;
    flxVec table;
    bn_rv rv;
  public:
    bn_obj_drv(const std::string name, const std::vector<std::string>& statesV, const std::vector<bn_rv_base*>& drvs, const flxVec& tableV, const tdouble roundoff_correct);
};

class bn_obj_detrel_and : public bn_node {
  private:
    const std::vector<std::string> states;
    bn_detRel_and drv;
  public:
    bn_obj_detrel_and(const std::string name, const std::vector<std::string>& statesV, const std::vector<bn_rv_base*>& drvs);
};

class bn_obj_dec : public bn_node {
  private:
    const std::vector<std::string> states;
    bn_dec dec;
  public:
    bn_obj_dec(const std::string name, const std::vector<std::string>& statesV);
};

class bn_obj_utility : public bn_node {
  private:
    const flxVec utilityVec;
    bn_utility ut;
  public:
    bn_obj_utility(const std::string name, const std::vector<bn_rv_base*>& drvs, const flxVec& utilityVecV);
};

class bn_obj_train_base : public bn_node {
  protected:
    std::string name;
    const tuint NstatesR;
    const tuint NstatesS;
    const flxVec ivalsR;
    const flxVec ivalsS;
    flxVec cptable1S;	// sum (start)
    flxVec cptable2S;	// sum (continue)
    std::valarray<bn_rv*> rV;
    std::valarray<bn_rv_base*> aV;
    std::valarray<bn_rv*> sV;
    std::vector<std::string> statesR;	// it's a simple numbering from 0 to ...
    std::vector<std::string> statesS;	// it's a simple numbering from 0 to ...
    const tuint Nstages;
    /**
    * @brief add two uniform distributions (for summation)
    * @param a1,a2 & b1,b2 lower (a) and upper (b) bounds of first and second uniform distribution
    * @param res discretized probability distribution (first interval is NOT the zero-value !!!)
    */
    void uniAdd(tdouble a1, tdouble b1, tdouble a2, tdouble b2, tdouble res[]) const;
    virtual void free_mem();
  public:
    bn_obj_train_base(const std::string& name, const flxVec& ivals, const flxVec& sivals, const tuint Nstages);
    virtual ~bn_obj_train_base();
};

class bn_obj_train : public bn_obj_train_base {
  private:
    flxVec ptableR;	// rain
    flxVec cptableR;	// rain
    flxVec cptableA;	// radar
  public:
    bn_obj_train(const std::string& name, const flxVec& ivals, const flxVec& ptable, const flxVec& cptable, const flxVec& rcptable, const flxVec& sivals, const tuint Nstages, const tdouble roundoff_correct);
};

class bn_obj_train_radar : public bn_obj_train_base {
  private:
    flxVec cptableRonA;	// radar
  public:
    bn_obj_train_radar(const std::string& name, const flxVec& ivals, const flxVec& cptable_RonA, const flxVec& sivals, const tuint Nstages, const tdouble roundoff_correct);
};



} // end of namespace 'flxbn_flxbn_manager_H'


#endif // flxbn_flxbn_obj_H

